'use strict';

var lodash = require('lodash');
var bluebird = require('bluebird');
var moleculer = require('moleculer');
var typeorm = require('typeorm');

/*! *****************************************************************************
Copyright (C) Microsoft. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
var Reflect;
(function (Reflect) {
    // Metadata Proposal
    // https://rbuckton.github.io/reflect-metadata/
    (function (factory) {
        var root = typeof global === "object" ? global :
            typeof self === "object" ? self :
                typeof this === "object" ? this :
                    Function("return this;")();
        var exporter = makeExporter(Reflect);
        if (typeof root.Reflect === "undefined") {
            root.Reflect = Reflect;
        }
        else {
            exporter = makeExporter(root.Reflect, exporter);
        }
        factory(exporter);
        function makeExporter(target, previous) {
            return function (key, value) {
                if (typeof target[key] !== "function") {
                    Object.defineProperty(target, key, { configurable: true, writable: true, value: value });
                }
                if (previous)
                    previous(key, value);
            };
        }
    })(function (exporter) {
        var hasOwn = Object.prototype.hasOwnProperty;
        // feature test for Symbol support
        var supportsSymbol = typeof Symbol === "function";
        var toPrimitiveSymbol = supportsSymbol && typeof Symbol.toPrimitive !== "undefined" ? Symbol.toPrimitive : "@@toPrimitive";
        var iteratorSymbol = supportsSymbol && typeof Symbol.iterator !== "undefined" ? Symbol.iterator : "@@iterator";
        var supportsCreate = typeof Object.create === "function"; // feature test for Object.create support
        var supportsProto = { __proto__: [] } instanceof Array; // feature test for __proto__ support
        var downLevel = !supportsCreate && !supportsProto;
        var HashMap = {
            // create an object in dictionary mode (a.k.a. "slow" mode in v8)
            create: supportsCreate
                ? function () { return MakeDictionary(Object.create(null)); }
                : supportsProto
                    ? function () { return MakeDictionary({ __proto__: null }); }
                    : function () { return MakeDictionary({}); },
            has: downLevel
                ? function (map, key) { return hasOwn.call(map, key); }
                : function (map, key) { return key in map; },
            get: downLevel
                ? function (map, key) { return hasOwn.call(map, key) ? map[key] : undefined; }
                : function (map, key) { return map[key]; },
        };
        // Load global or shim versions of Map, Set, and WeakMap
        var functionPrototype = Object.getPrototypeOf(Function);
        var usePolyfill = typeof process === "object" && process.env && process.env["REFLECT_METADATA_USE_MAP_POLYFILL"] === "true";
        var _Map = !usePolyfill && typeof Map === "function" && typeof Map.prototype.entries === "function" ? Map : CreateMapPolyfill();
        var _Set = !usePolyfill && typeof Set === "function" && typeof Set.prototype.entries === "function" ? Set : CreateSetPolyfill();
        var _WeakMap = !usePolyfill && typeof WeakMap === "function" ? WeakMap : CreateWeakMapPolyfill();
        // [[Metadata]] internal slot
        // https://rbuckton.github.io/reflect-metadata/#ordinary-object-internal-methods-and-internal-slots
        var Metadata = new _WeakMap();
        /**
         * Applies a set of decorators to a property of a target object.
         * @param decorators An array of decorators.
         * @param target The target object.
         * @param propertyKey (Optional) The property key to decorate.
         * @param attributes (Optional) The property descriptor for the target key.
         * @remarks Decorators are applied in reverse order.
         * @example
         *
         *     class Example {
         *         // property declarations are not part of ES6, though they are valid in TypeScript:
         *         // static staticProperty;
         *         // property;
         *
         *         constructor(p) { }
         *         static staticMethod(p) { }
         *         method(p) { }
         *     }
         *
         *     // constructor
         *     Example = Reflect.decorate(decoratorsArray, Example);
         *
         *     // property (on constructor)
         *     Reflect.decorate(decoratorsArray, Example, "staticProperty");
         *
         *     // property (on prototype)
         *     Reflect.decorate(decoratorsArray, Example.prototype, "property");
         *
         *     // method (on constructor)
         *     Object.defineProperty(Example, "staticMethod",
         *         Reflect.decorate(decoratorsArray, Example, "staticMethod",
         *             Object.getOwnPropertyDescriptor(Example, "staticMethod")));
         *
         *     // method (on prototype)
         *     Object.defineProperty(Example.prototype, "method",
         *         Reflect.decorate(decoratorsArray, Example.prototype, "method",
         *             Object.getOwnPropertyDescriptor(Example.prototype, "method")));
         *
         */
        function decorate(decorators, target, propertyKey, attributes) {
            if (!IsUndefined(propertyKey)) {
                if (!IsArray(decorators))
                    throw new TypeError();
                if (!IsObject(target))
                    throw new TypeError();
                if (!IsObject(attributes) && !IsUndefined(attributes) && !IsNull(attributes))
                    throw new TypeError();
                if (IsNull(attributes))
                    attributes = undefined;
                propertyKey = ToPropertyKey(propertyKey);
                return DecorateProperty(decorators, target, propertyKey, attributes);
            }
            else {
                if (!IsArray(decorators))
                    throw new TypeError();
                if (!IsConstructor(target))
                    throw new TypeError();
                return DecorateConstructor(decorators, target);
            }
        }
        exporter("decorate", decorate);
        // 4.1.2 Reflect.metadata(metadataKey, metadataValue)
        // https://rbuckton.github.io/reflect-metadata/#reflect.metadata
        /**
         * A default metadata decorator factory that can be used on a class, class member, or parameter.
         * @param metadataKey The key for the metadata entry.
         * @param metadataValue The value for the metadata entry.
         * @returns A decorator function.
         * @remarks
         * If `metadataKey` is already defined for the target and target key, the
         * metadataValue for that key will be overwritten.
         * @example
         *
         *     // constructor
         *     @Reflect.metadata(key, value)
         *     class Example {
         *     }
         *
         *     // property (on constructor, TypeScript only)
         *     class Example {
         *         @Reflect.metadata(key, value)
         *         static staticProperty;
         *     }
         *
         *     // property (on prototype, TypeScript only)
         *     class Example {
         *         @Reflect.metadata(key, value)
         *         property;
         *     }
         *
         *     // method (on constructor)
         *     class Example {
         *         @Reflect.metadata(key, value)
         *         static staticMethod() { }
         *     }
         *
         *     // method (on prototype)
         *     class Example {
         *         @Reflect.metadata(key, value)
         *         method() { }
         *     }
         *
         */
        function metadata(metadataKey, metadataValue) {
            function decorator(target, propertyKey) {
                if (!IsObject(target))
                    throw new TypeError();
                if (!IsUndefined(propertyKey) && !IsPropertyKey(propertyKey))
                    throw new TypeError();
                OrdinaryDefineOwnMetadata(metadataKey, metadataValue, target, propertyKey);
            }
            return decorator;
        }
        exporter("metadata", metadata);
        /**
         * Define a unique metadata entry on the target.
         * @param metadataKey A key used to store and retrieve metadata.
         * @param metadataValue A value that contains attached metadata.
         * @param target The target object on which to define metadata.
         * @param propertyKey (Optional) The property key for the target.
         * @example
         *
         *     class Example {
         *         // property declarations are not part of ES6, though they are valid in TypeScript:
         *         // static staticProperty;
         *         // property;
         *
         *         constructor(p) { }
         *         static staticMethod(p) { }
         *         method(p) { }
         *     }
         *
         *     // constructor
         *     Reflect.defineMetadata("custom:annotation", options, Example);
         *
         *     // property (on constructor)
         *     Reflect.defineMetadata("custom:annotation", options, Example, "staticProperty");
         *
         *     // property (on prototype)
         *     Reflect.defineMetadata("custom:annotation", options, Example.prototype, "property");
         *
         *     // method (on constructor)
         *     Reflect.defineMetadata("custom:annotation", options, Example, "staticMethod");
         *
         *     // method (on prototype)
         *     Reflect.defineMetadata("custom:annotation", options, Example.prototype, "method");
         *
         *     // decorator factory as metadata-producing annotation.
         *     function MyAnnotation(options): Decorator {
         *         return (target, key?) => Reflect.defineMetadata("custom:annotation", options, target, key);
         *     }
         *
         */
        function defineMetadata(metadataKey, metadataValue, target, propertyKey) {
            if (!IsObject(target))
                throw new TypeError();
            if (!IsUndefined(propertyKey))
                propertyKey = ToPropertyKey(propertyKey);
            return OrdinaryDefineOwnMetadata(metadataKey, metadataValue, target, propertyKey);
        }
        exporter("defineMetadata", defineMetadata);
        /**
         * Gets a value indicating whether the target object or its prototype chain has the provided metadata key defined.
         * @param metadataKey A key used to store and retrieve metadata.
         * @param target The target object on which the metadata is defined.
         * @param propertyKey (Optional) The property key for the target.
         * @returns `true` if the metadata key was defined on the target object or its prototype chain; otherwise, `false`.
         * @example
         *
         *     class Example {
         *         // property declarations are not part of ES6, though they are valid in TypeScript:
         *         // static staticProperty;
         *         // property;
         *
         *         constructor(p) { }
         *         static staticMethod(p) { }
         *         method(p) { }
         *     }
         *
         *     // constructor
         *     result = Reflect.hasMetadata("custom:annotation", Example);
         *
         *     // property (on constructor)
         *     result = Reflect.hasMetadata("custom:annotation", Example, "staticProperty");
         *
         *     // property (on prototype)
         *     result = Reflect.hasMetadata("custom:annotation", Example.prototype, "property");
         *
         *     // method (on constructor)
         *     result = Reflect.hasMetadata("custom:annotation", Example, "staticMethod");
         *
         *     // method (on prototype)
         *     result = Reflect.hasMetadata("custom:annotation", Example.prototype, "method");
         *
         */
        function hasMetadata(metadataKey, target, propertyKey) {
            if (!IsObject(target))
                throw new TypeError();
            if (!IsUndefined(propertyKey))
                propertyKey = ToPropertyKey(propertyKey);
            return OrdinaryHasMetadata(metadataKey, target, propertyKey);
        }
        exporter("hasMetadata", hasMetadata);
        /**
         * Gets a value indicating whether the target object has the provided metadata key defined.
         * @param metadataKey A key used to store and retrieve metadata.
         * @param target The target object on which the metadata is defined.
         * @param propertyKey (Optional) The property key for the target.
         * @returns `true` if the metadata key was defined on the target object; otherwise, `false`.
         * @example
         *
         *     class Example {
         *         // property declarations are not part of ES6, though they are valid in TypeScript:
         *         // static staticProperty;
         *         // property;
         *
         *         constructor(p) { }
         *         static staticMethod(p) { }
         *         method(p) { }
         *     }
         *
         *     // constructor
         *     result = Reflect.hasOwnMetadata("custom:annotation", Example);
         *
         *     // property (on constructor)
         *     result = Reflect.hasOwnMetadata("custom:annotation", Example, "staticProperty");
         *
         *     // property (on prototype)
         *     result = Reflect.hasOwnMetadata("custom:annotation", Example.prototype, "property");
         *
         *     // method (on constructor)
         *     result = Reflect.hasOwnMetadata("custom:annotation", Example, "staticMethod");
         *
         *     // method (on prototype)
         *     result = Reflect.hasOwnMetadata("custom:annotation", Example.prototype, "method");
         *
         */
        function hasOwnMetadata(metadataKey, target, propertyKey) {
            if (!IsObject(target))
                throw new TypeError();
            if (!IsUndefined(propertyKey))
                propertyKey = ToPropertyKey(propertyKey);
            return OrdinaryHasOwnMetadata(metadataKey, target, propertyKey);
        }
        exporter("hasOwnMetadata", hasOwnMetadata);
        /**
         * Gets the metadata value for the provided metadata key on the target object or its prototype chain.
         * @param metadataKey A key used to store and retrieve metadata.
         * @param target The target object on which the metadata is defined.
         * @param propertyKey (Optional) The property key for the target.
         * @returns The metadata value for the metadata key if found; otherwise, `undefined`.
         * @example
         *
         *     class Example {
         *         // property declarations are not part of ES6, though they are valid in TypeScript:
         *         // static staticProperty;
         *         // property;
         *
         *         constructor(p) { }
         *         static staticMethod(p) { }
         *         method(p) { }
         *     }
         *
         *     // constructor
         *     result = Reflect.getMetadata("custom:annotation", Example);
         *
         *     // property (on constructor)
         *     result = Reflect.getMetadata("custom:annotation", Example, "staticProperty");
         *
         *     // property (on prototype)
         *     result = Reflect.getMetadata("custom:annotation", Example.prototype, "property");
         *
         *     // method (on constructor)
         *     result = Reflect.getMetadata("custom:annotation", Example, "staticMethod");
         *
         *     // method (on prototype)
         *     result = Reflect.getMetadata("custom:annotation", Example.prototype, "method");
         *
         */
        function getMetadata(metadataKey, target, propertyKey) {
            if (!IsObject(target))
                throw new TypeError();
            if (!IsUndefined(propertyKey))
                propertyKey = ToPropertyKey(propertyKey);
            return OrdinaryGetMetadata(metadataKey, target, propertyKey);
        }
        exporter("getMetadata", getMetadata);
        /**
         * Gets the metadata value for the provided metadata key on the target object.
         * @param metadataKey A key used to store and retrieve metadata.
         * @param target The target object on which the metadata is defined.
         * @param propertyKey (Optional) The property key for the target.
         * @returns The metadata value for the metadata key if found; otherwise, `undefined`.
         * @example
         *
         *     class Example {
         *         // property declarations are not part of ES6, though they are valid in TypeScript:
         *         // static staticProperty;
         *         // property;
         *
         *         constructor(p) { }
         *         static staticMethod(p) { }
         *         method(p) { }
         *     }
         *
         *     // constructor
         *     result = Reflect.getOwnMetadata("custom:annotation", Example);
         *
         *     // property (on constructor)
         *     result = Reflect.getOwnMetadata("custom:annotation", Example, "staticProperty");
         *
         *     // property (on prototype)
         *     result = Reflect.getOwnMetadata("custom:annotation", Example.prototype, "property");
         *
         *     // method (on constructor)
         *     result = Reflect.getOwnMetadata("custom:annotation", Example, "staticMethod");
         *
         *     // method (on prototype)
         *     result = Reflect.getOwnMetadata("custom:annotation", Example.prototype, "method");
         *
         */
        function getOwnMetadata(metadataKey, target, propertyKey) {
            if (!IsObject(target))
                throw new TypeError();
            if (!IsUndefined(propertyKey))
                propertyKey = ToPropertyKey(propertyKey);
            return OrdinaryGetOwnMetadata(metadataKey, target, propertyKey);
        }
        exporter("getOwnMetadata", getOwnMetadata);
        /**
         * Gets the metadata keys defined on the target object or its prototype chain.
         * @param target The target object on which the metadata is defined.
         * @param propertyKey (Optional) The property key for the target.
         * @returns An array of unique metadata keys.
         * @example
         *
         *     class Example {
         *         // property declarations are not part of ES6, though they are valid in TypeScript:
         *         // static staticProperty;
         *         // property;
         *
         *         constructor(p) { }
         *         static staticMethod(p) { }
         *         method(p) { }
         *     }
         *
         *     // constructor
         *     result = Reflect.getMetadataKeys(Example);
         *
         *     // property (on constructor)
         *     result = Reflect.getMetadataKeys(Example, "staticProperty");
         *
         *     // property (on prototype)
         *     result = Reflect.getMetadataKeys(Example.prototype, "property");
         *
         *     // method (on constructor)
         *     result = Reflect.getMetadataKeys(Example, "staticMethod");
         *
         *     // method (on prototype)
         *     result = Reflect.getMetadataKeys(Example.prototype, "method");
         *
         */
        function getMetadataKeys(target, propertyKey) {
            if (!IsObject(target))
                throw new TypeError();
            if (!IsUndefined(propertyKey))
                propertyKey = ToPropertyKey(propertyKey);
            return OrdinaryMetadataKeys(target, propertyKey);
        }
        exporter("getMetadataKeys", getMetadataKeys);
        /**
         * Gets the unique metadata keys defined on the target object.
         * @param target The target object on which the metadata is defined.
         * @param propertyKey (Optional) The property key for the target.
         * @returns An array of unique metadata keys.
         * @example
         *
         *     class Example {
         *         // property declarations are not part of ES6, though they are valid in TypeScript:
         *         // static staticProperty;
         *         // property;
         *
         *         constructor(p) { }
         *         static staticMethod(p) { }
         *         method(p) { }
         *     }
         *
         *     // constructor
         *     result = Reflect.getOwnMetadataKeys(Example);
         *
         *     // property (on constructor)
         *     result = Reflect.getOwnMetadataKeys(Example, "staticProperty");
         *
         *     // property (on prototype)
         *     result = Reflect.getOwnMetadataKeys(Example.prototype, "property");
         *
         *     // method (on constructor)
         *     result = Reflect.getOwnMetadataKeys(Example, "staticMethod");
         *
         *     // method (on prototype)
         *     result = Reflect.getOwnMetadataKeys(Example.prototype, "method");
         *
         */
        function getOwnMetadataKeys(target, propertyKey) {
            if (!IsObject(target))
                throw new TypeError();
            if (!IsUndefined(propertyKey))
                propertyKey = ToPropertyKey(propertyKey);
            return OrdinaryOwnMetadataKeys(target, propertyKey);
        }
        exporter("getOwnMetadataKeys", getOwnMetadataKeys);
        /**
         * Deletes the metadata entry from the target object with the provided key.
         * @param metadataKey A key used to store and retrieve metadata.
         * @param target The target object on which the metadata is defined.
         * @param propertyKey (Optional) The property key for the target.
         * @returns `true` if the metadata entry was found and deleted; otherwise, false.
         * @example
         *
         *     class Example {
         *         // property declarations are not part of ES6, though they are valid in TypeScript:
         *         // static staticProperty;
         *         // property;
         *
         *         constructor(p) { }
         *         static staticMethod(p) { }
         *         method(p) { }
         *     }
         *
         *     // constructor
         *     result = Reflect.deleteMetadata("custom:annotation", Example);
         *
         *     // property (on constructor)
         *     result = Reflect.deleteMetadata("custom:annotation", Example, "staticProperty");
         *
         *     // property (on prototype)
         *     result = Reflect.deleteMetadata("custom:annotation", Example.prototype, "property");
         *
         *     // method (on constructor)
         *     result = Reflect.deleteMetadata("custom:annotation", Example, "staticMethod");
         *
         *     // method (on prototype)
         *     result = Reflect.deleteMetadata("custom:annotation", Example.prototype, "method");
         *
         */
        function deleteMetadata(metadataKey, target, propertyKey) {
            if (!IsObject(target))
                throw new TypeError();
            if (!IsUndefined(propertyKey))
                propertyKey = ToPropertyKey(propertyKey);
            var metadataMap = GetOrCreateMetadataMap(target, propertyKey, /*Create*/ false);
            if (IsUndefined(metadataMap))
                return false;
            if (!metadataMap.delete(metadataKey))
                return false;
            if (metadataMap.size > 0)
                return true;
            var targetMetadata = Metadata.get(target);
            targetMetadata.delete(propertyKey);
            if (targetMetadata.size > 0)
                return true;
            Metadata.delete(target);
            return true;
        }
        exporter("deleteMetadata", deleteMetadata);
        function DecorateConstructor(decorators, target) {
            for (var i = decorators.length - 1; i >= 0; --i) {
                var decorator = decorators[i];
                var decorated = decorator(target);
                if (!IsUndefined(decorated) && !IsNull(decorated)) {
                    if (!IsConstructor(decorated))
                        throw new TypeError();
                    target = decorated;
                }
            }
            return target;
        }
        function DecorateProperty(decorators, target, propertyKey, descriptor) {
            for (var i = decorators.length - 1; i >= 0; --i) {
                var decorator = decorators[i];
                var decorated = decorator(target, propertyKey, descriptor);
                if (!IsUndefined(decorated) && !IsNull(decorated)) {
                    if (!IsObject(decorated))
                        throw new TypeError();
                    descriptor = decorated;
                }
            }
            return descriptor;
        }
        function GetOrCreateMetadataMap(O, P, Create) {
            var targetMetadata = Metadata.get(O);
            if (IsUndefined(targetMetadata)) {
                if (!Create)
                    return undefined;
                targetMetadata = new _Map();
                Metadata.set(O, targetMetadata);
            }
            var metadataMap = targetMetadata.get(P);
            if (IsUndefined(metadataMap)) {
                if (!Create)
                    return undefined;
                metadataMap = new _Map();
                targetMetadata.set(P, metadataMap);
            }
            return metadataMap;
        }
        // 3.1.1.1 OrdinaryHasMetadata(MetadataKey, O, P)
        // https://rbuckton.github.io/reflect-metadata/#ordinaryhasmetadata
        function OrdinaryHasMetadata(MetadataKey, O, P) {
            var hasOwn = OrdinaryHasOwnMetadata(MetadataKey, O, P);
            if (hasOwn)
                return true;
            var parent = OrdinaryGetPrototypeOf(O);
            if (!IsNull(parent))
                return OrdinaryHasMetadata(MetadataKey, parent, P);
            return false;
        }
        // 3.1.2.1 OrdinaryHasOwnMetadata(MetadataKey, O, P)
        // https://rbuckton.github.io/reflect-metadata/#ordinaryhasownmetadata
        function OrdinaryHasOwnMetadata(MetadataKey, O, P) {
            var metadataMap = GetOrCreateMetadataMap(O, P, /*Create*/ false);
            if (IsUndefined(metadataMap))
                return false;
            return ToBoolean(metadataMap.has(MetadataKey));
        }
        // 3.1.3.1 OrdinaryGetMetadata(MetadataKey, O, P)
        // https://rbuckton.github.io/reflect-metadata/#ordinarygetmetadata
        function OrdinaryGetMetadata(MetadataKey, O, P) {
            var hasOwn = OrdinaryHasOwnMetadata(MetadataKey, O, P);
            if (hasOwn)
                return OrdinaryGetOwnMetadata(MetadataKey, O, P);
            var parent = OrdinaryGetPrototypeOf(O);
            if (!IsNull(parent))
                return OrdinaryGetMetadata(MetadataKey, parent, P);
            return undefined;
        }
        // 3.1.4.1 OrdinaryGetOwnMetadata(MetadataKey, O, P)
        // https://rbuckton.github.io/reflect-metadata/#ordinarygetownmetadata
        function OrdinaryGetOwnMetadata(MetadataKey, O, P) {
            var metadataMap = GetOrCreateMetadataMap(O, P, /*Create*/ false);
            if (IsUndefined(metadataMap))
                return undefined;
            return metadataMap.get(MetadataKey);
        }
        // 3.1.5.1 OrdinaryDefineOwnMetadata(MetadataKey, MetadataValue, O, P)
        // https://rbuckton.github.io/reflect-metadata/#ordinarydefineownmetadata
        function OrdinaryDefineOwnMetadata(MetadataKey, MetadataValue, O, P) {
            var metadataMap = GetOrCreateMetadataMap(O, P, /*Create*/ true);
            metadataMap.set(MetadataKey, MetadataValue);
        }
        // 3.1.6.1 OrdinaryMetadataKeys(O, P)
        // https://rbuckton.github.io/reflect-metadata/#ordinarymetadatakeys
        function OrdinaryMetadataKeys(O, P) {
            var ownKeys = OrdinaryOwnMetadataKeys(O, P);
            var parent = OrdinaryGetPrototypeOf(O);
            if (parent === null)
                return ownKeys;
            var parentKeys = OrdinaryMetadataKeys(parent, P);
            if (parentKeys.length <= 0)
                return ownKeys;
            if (ownKeys.length <= 0)
                return parentKeys;
            var set = new _Set();
            var keys = [];
            for (var _i = 0, ownKeys_1 = ownKeys; _i < ownKeys_1.length; _i++) {
                var key = ownKeys_1[_i];
                var hasKey = set.has(key);
                if (!hasKey) {
                    set.add(key);
                    keys.push(key);
                }
            }
            for (var _a = 0, parentKeys_1 = parentKeys; _a < parentKeys_1.length; _a++) {
                var key = parentKeys_1[_a];
                var hasKey = set.has(key);
                if (!hasKey) {
                    set.add(key);
                    keys.push(key);
                }
            }
            return keys;
        }
        // 3.1.7.1 OrdinaryOwnMetadataKeys(O, P)
        // https://rbuckton.github.io/reflect-metadata/#ordinaryownmetadatakeys
        function OrdinaryOwnMetadataKeys(O, P) {
            var keys = [];
            var metadataMap = GetOrCreateMetadataMap(O, P, /*Create*/ false);
            if (IsUndefined(metadataMap))
                return keys;
            var keysObj = metadataMap.keys();
            var iterator = GetIterator(keysObj);
            var k = 0;
            while (true) {
                var next = IteratorStep(iterator);
                if (!next) {
                    keys.length = k;
                    return keys;
                }
                var nextValue = IteratorValue(next);
                try {
                    keys[k] = nextValue;
                }
                catch (e) {
                    try {
                        IteratorClose(iterator);
                    }
                    finally {
                        throw e;
                    }
                }
                k++;
            }
        }
        // 6 ECMAScript Data Typ0es and Values
        // https://tc39.github.io/ecma262/#sec-ecmascript-data-types-and-values
        function Type(x) {
            if (x === null)
                return 1 /* Null */;
            switch (typeof x) {
                case "undefined": return 0 /* Undefined */;
                case "boolean": return 2 /* Boolean */;
                case "string": return 3 /* String */;
                case "symbol": return 4 /* Symbol */;
                case "number": return 5 /* Number */;
                case "object": return x === null ? 1 /* Null */ : 6 /* Object */;
                default: return 6 /* Object */;
            }
        }
        // 6.1.1 The Undefined Type
        // https://tc39.github.io/ecma262/#sec-ecmascript-language-types-undefined-type
        function IsUndefined(x) {
            return x === undefined;
        }
        // 6.1.2 The Null Type
        // https://tc39.github.io/ecma262/#sec-ecmascript-language-types-null-type
        function IsNull(x) {
            return x === null;
        }
        // 6.1.5 The Symbol Type
        // https://tc39.github.io/ecma262/#sec-ecmascript-language-types-symbol-type
        function IsSymbol(x) {
            return typeof x === "symbol";
        }
        // 6.1.7 The Object Type
        // https://tc39.github.io/ecma262/#sec-object-type
        function IsObject(x) {
            return typeof x === "object" ? x !== null : typeof x === "function";
        }
        // 7.1 Type Conversion
        // https://tc39.github.io/ecma262/#sec-type-conversion
        // 7.1.1 ToPrimitive(input [, PreferredType])
        // https://tc39.github.io/ecma262/#sec-toprimitive
        function ToPrimitive(input, PreferredType) {
            switch (Type(input)) {
                case 0 /* Undefined */: return input;
                case 1 /* Null */: return input;
                case 2 /* Boolean */: return input;
                case 3 /* String */: return input;
                case 4 /* Symbol */: return input;
                case 5 /* Number */: return input;
            }
            var hint = PreferredType === 3 /* String */ ? "string" : PreferredType === 5 /* Number */ ? "number" : "default";
            var exoticToPrim = GetMethod(input, toPrimitiveSymbol);
            if (exoticToPrim !== undefined) {
                var result = exoticToPrim.call(input, hint);
                if (IsObject(result))
                    throw new TypeError();
                return result;
            }
            return OrdinaryToPrimitive(input, hint === "default" ? "number" : hint);
        }
        // 7.1.1.1 OrdinaryToPrimitive(O, hint)
        // https://tc39.github.io/ecma262/#sec-ordinarytoprimitive
        function OrdinaryToPrimitive(O, hint) {
            if (hint === "string") {
                var toString_1 = O.toString;
                if (IsCallable(toString_1)) {
                    var result = toString_1.call(O);
                    if (!IsObject(result))
                        return result;
                }
                var valueOf = O.valueOf;
                if (IsCallable(valueOf)) {
                    var result = valueOf.call(O);
                    if (!IsObject(result))
                        return result;
                }
            }
            else {
                var valueOf = O.valueOf;
                if (IsCallable(valueOf)) {
                    var result = valueOf.call(O);
                    if (!IsObject(result))
                        return result;
                }
                var toString_2 = O.toString;
                if (IsCallable(toString_2)) {
                    var result = toString_2.call(O);
                    if (!IsObject(result))
                        return result;
                }
            }
            throw new TypeError();
        }
        // 7.1.2 ToBoolean(argument)
        // https://tc39.github.io/ecma262/2016/#sec-toboolean
        function ToBoolean(argument) {
            return !!argument;
        }
        // 7.1.12 ToString(argument)
        // https://tc39.github.io/ecma262/#sec-tostring
        function ToString(argument) {
            return "" + argument;
        }
        // 7.1.14 ToPropertyKey(argument)
        // https://tc39.github.io/ecma262/#sec-topropertykey
        function ToPropertyKey(argument) {
            var key = ToPrimitive(argument, 3 /* String */);
            if (IsSymbol(key))
                return key;
            return ToString(key);
        }
        // 7.2 Testing and Comparison Operations
        // https://tc39.github.io/ecma262/#sec-testing-and-comparison-operations
        // 7.2.2 IsArray(argument)
        // https://tc39.github.io/ecma262/#sec-isarray
        function IsArray(argument) {
            return Array.isArray
                ? Array.isArray(argument)
                : argument instanceof Object
                    ? argument instanceof Array
                    : Object.prototype.toString.call(argument) === "[object Array]";
        }
        // 7.2.3 IsCallable(argument)
        // https://tc39.github.io/ecma262/#sec-iscallable
        function IsCallable(argument) {
            // NOTE: This is an approximation as we cannot check for [[Call]] internal method.
            return typeof argument === "function";
        }
        // 7.2.4 IsConstructor(argument)
        // https://tc39.github.io/ecma262/#sec-isconstructor
        function IsConstructor(argument) {
            // NOTE: This is an approximation as we cannot check for [[Construct]] internal method.
            return typeof argument === "function";
        }
        // 7.2.7 IsPropertyKey(argument)
        // https://tc39.github.io/ecma262/#sec-ispropertykey
        function IsPropertyKey(argument) {
            switch (Type(argument)) {
                case 3 /* String */: return true;
                case 4 /* Symbol */: return true;
                default: return false;
            }
        }
        // 7.3 Operations on Objects
        // https://tc39.github.io/ecma262/#sec-operations-on-objects
        // 7.3.9 GetMethod(V, P)
        // https://tc39.github.io/ecma262/#sec-getmethod
        function GetMethod(V, P) {
            var func = V[P];
            if (func === undefined || func === null)
                return undefined;
            if (!IsCallable(func))
                throw new TypeError();
            return func;
        }
        // 7.4 Operations on Iterator Objects
        // https://tc39.github.io/ecma262/#sec-operations-on-iterator-objects
        function GetIterator(obj) {
            var method = GetMethod(obj, iteratorSymbol);
            if (!IsCallable(method))
                throw new TypeError(); // from Call
            var iterator = method.call(obj);
            if (!IsObject(iterator))
                throw new TypeError();
            return iterator;
        }
        // 7.4.4 IteratorValue(iterResult)
        // https://tc39.github.io/ecma262/2016/#sec-iteratorvalue
        function IteratorValue(iterResult) {
            return iterResult.value;
        }
        // 7.4.5 IteratorStep(iterator)
        // https://tc39.github.io/ecma262/#sec-iteratorstep
        function IteratorStep(iterator) {
            var result = iterator.next();
            return result.done ? false : result;
        }
        // 7.4.6 IteratorClose(iterator, completion)
        // https://tc39.github.io/ecma262/#sec-iteratorclose
        function IteratorClose(iterator) {
            var f = iterator["return"];
            if (f)
                f.call(iterator);
        }
        // 9.1 Ordinary Object Internal Methods and Internal Slots
        // https://tc39.github.io/ecma262/#sec-ordinary-object-internal-methods-and-internal-slots
        // 9.1.1.1 OrdinaryGetPrototypeOf(O)
        // https://tc39.github.io/ecma262/#sec-ordinarygetprototypeof
        function OrdinaryGetPrototypeOf(O) {
            var proto = Object.getPrototypeOf(O);
            if (typeof O !== "function" || O === functionPrototype)
                return proto;
            // TypeScript doesn't set __proto__ in ES5, as it's non-standard.
            // Try to determine the superclass constructor. Compatible implementations
            // must either set __proto__ on a subclass constructor to the superclass constructor,
            // or ensure each class has a valid `constructor` property on its prototype that
            // points back to the constructor.
            // If this is not the same as Function.[[Prototype]], then this is definately inherited.
            // This is the case when in ES6 or when using __proto__ in a compatible browser.
            if (proto !== functionPrototype)
                return proto;
            // If the super prototype is Object.prototype, null, or undefined, then we cannot determine the heritage.
            var prototype = O.prototype;
            var prototypeProto = prototype && Object.getPrototypeOf(prototype);
            if (prototypeProto == null || prototypeProto === Object.prototype)
                return proto;
            // If the constructor was not a function, then we cannot determine the heritage.
            var constructor = prototypeProto.constructor;
            if (typeof constructor !== "function")
                return proto;
            // If we have some kind of self-reference, then we cannot determine the heritage.
            if (constructor === O)
                return proto;
            // we have a pretty good guess at the heritage.
            return constructor;
        }
        // naive Map shim
        function CreateMapPolyfill() {
            var cacheSentinel = {};
            var arraySentinel = [];
            var MapIterator = /** @class */ (function () {
                function MapIterator(keys, values, selector) {
                    this._index = 0;
                    this._keys = keys;
                    this._values = values;
                    this._selector = selector;
                }
                MapIterator.prototype["@@iterator"] = function () { return this; };
                MapIterator.prototype[iteratorSymbol] = function () { return this; };
                MapIterator.prototype.next = function () {
                    var index = this._index;
                    if (index >= 0 && index < this._keys.length) {
                        var result = this._selector(this._keys[index], this._values[index]);
                        if (index + 1 >= this._keys.length) {
                            this._index = -1;
                            this._keys = arraySentinel;
                            this._values = arraySentinel;
                        }
                        else {
                            this._index++;
                        }
                        return { value: result, done: false };
                    }
                    return { value: undefined, done: true };
                };
                MapIterator.prototype.throw = function (error) {
                    if (this._index >= 0) {
                        this._index = -1;
                        this._keys = arraySentinel;
                        this._values = arraySentinel;
                    }
                    throw error;
                };
                MapIterator.prototype.return = function (value) {
                    if (this._index >= 0) {
                        this._index = -1;
                        this._keys = arraySentinel;
                        this._values = arraySentinel;
                    }
                    return { value: value, done: true };
                };
                return MapIterator;
            }());
            return /** @class */ (function () {
                function Map() {
                    this._keys = [];
                    this._values = [];
                    this._cacheKey = cacheSentinel;
                    this._cacheIndex = -2;
                }
                Object.defineProperty(Map.prototype, "size", {
                    get: function () { return this._keys.length; },
                    enumerable: true,
                    configurable: true
                });
                Map.prototype.has = function (key) { return this._find(key, /*insert*/ false) >= 0; };
                Map.prototype.get = function (key) {
                    var index = this._find(key, /*insert*/ false);
                    return index >= 0 ? this._values[index] : undefined;
                };
                Map.prototype.set = function (key, value) {
                    var index = this._find(key, /*insert*/ true);
                    this._values[index] = value;
                    return this;
                };
                Map.prototype.delete = function (key) {
                    var index = this._find(key, /*insert*/ false);
                    if (index >= 0) {
                        var size = this._keys.length;
                        for (var i = index + 1; i < size; i++) {
                            this._keys[i - 1] = this._keys[i];
                            this._values[i - 1] = this._values[i];
                        }
                        this._keys.length--;
                        this._values.length--;
                        if (key === this._cacheKey) {
                            this._cacheKey = cacheSentinel;
                            this._cacheIndex = -2;
                        }
                        return true;
                    }
                    return false;
                };
                Map.prototype.clear = function () {
                    this._keys.length = 0;
                    this._values.length = 0;
                    this._cacheKey = cacheSentinel;
                    this._cacheIndex = -2;
                };
                Map.prototype.keys = function () { return new MapIterator(this._keys, this._values, getKey); };
                Map.prototype.values = function () { return new MapIterator(this._keys, this._values, getValue); };
                Map.prototype.entries = function () { return new MapIterator(this._keys, this._values, getEntry); };
                Map.prototype["@@iterator"] = function () { return this.entries(); };
                Map.prototype[iteratorSymbol] = function () { return this.entries(); };
                Map.prototype._find = function (key, insert) {
                    if (this._cacheKey !== key) {
                        this._cacheIndex = this._keys.indexOf(this._cacheKey = key);
                    }
                    if (this._cacheIndex < 0 && insert) {
                        this._cacheIndex = this._keys.length;
                        this._keys.push(key);
                        this._values.push(undefined);
                    }
                    return this._cacheIndex;
                };
                return Map;
            }());
            function getKey(key, _) {
                return key;
            }
            function getValue(_, value) {
                return value;
            }
            function getEntry(key, value) {
                return [key, value];
            }
        }
        // naive Set shim
        function CreateSetPolyfill() {
            return /** @class */ (function () {
                function Set() {
                    this._map = new _Map();
                }
                Object.defineProperty(Set.prototype, "size", {
                    get: function () { return this._map.size; },
                    enumerable: true,
                    configurable: true
                });
                Set.prototype.has = function (value) { return this._map.has(value); };
                Set.prototype.add = function (value) { return this._map.set(value, value), this; };
                Set.prototype.delete = function (value) { return this._map.delete(value); };
                Set.prototype.clear = function () { this._map.clear(); };
                Set.prototype.keys = function () { return this._map.keys(); };
                Set.prototype.values = function () { return this._map.values(); };
                Set.prototype.entries = function () { return this._map.entries(); };
                Set.prototype["@@iterator"] = function () { return this.keys(); };
                Set.prototype[iteratorSymbol] = function () { return this.keys(); };
                return Set;
            }());
        }
        // naive WeakMap shim
        function CreateWeakMapPolyfill() {
            var UUID_SIZE = 16;
            var keys = HashMap.create();
            var rootKey = CreateUniqueKey();
            return /** @class */ (function () {
                function WeakMap() {
                    this._key = CreateUniqueKey();
                }
                WeakMap.prototype.has = function (target) {
                    var table = GetOrCreateWeakMapTable(target, /*create*/ false);
                    return table !== undefined ? HashMap.has(table, this._key) : false;
                };
                WeakMap.prototype.get = function (target) {
                    var table = GetOrCreateWeakMapTable(target, /*create*/ false);
                    return table !== undefined ? HashMap.get(table, this._key) : undefined;
                };
                WeakMap.prototype.set = function (target, value) {
                    var table = GetOrCreateWeakMapTable(target, /*create*/ true);
                    table[this._key] = value;
                    return this;
                };
                WeakMap.prototype.delete = function (target) {
                    var table = GetOrCreateWeakMapTable(target, /*create*/ false);
                    return table !== undefined ? delete table[this._key] : false;
                };
                WeakMap.prototype.clear = function () {
                    // NOTE: not a real clear, just makes the previous data unreachable
                    this._key = CreateUniqueKey();
                };
                return WeakMap;
            }());
            function CreateUniqueKey() {
                var key;
                do
                    key = "@@WeakMap@@" + CreateUUID();
                while (HashMap.has(keys, key));
                keys[key] = true;
                return key;
            }
            function GetOrCreateWeakMapTable(target, create) {
                if (!hasOwn.call(target, rootKey)) {
                    if (!create)
                        return undefined;
                    Object.defineProperty(target, rootKey, { value: HashMap.create() });
                }
                return target[rootKey];
            }
            function FillRandomBytes(buffer, size) {
                for (var i = 0; i < size; ++i)
                    buffer[i] = Math.random() * 0xff | 0;
                return buffer;
            }
            function GenRandomBytes(size) {
                if (typeof Uint8Array === "function") {
                    if (typeof crypto !== "undefined")
                        return crypto.getRandomValues(new Uint8Array(size));
                    if (typeof msCrypto !== "undefined")
                        return msCrypto.getRandomValues(new Uint8Array(size));
                    return FillRandomBytes(new Uint8Array(size), size);
                }
                return FillRandomBytes(new Array(size), size);
            }
            function CreateUUID() {
                var data = GenRandomBytes(UUID_SIZE);
                // mark as random - RFC 4122 § 4.4
                data[6] = data[6] & 0x4f | 0x40;
                data[8] = data[8] & 0xbf | 0x80;
                var result = "";
                for (var offset = 0; offset < UUID_SIZE; ++offset) {
                    var byte = data[offset];
                    if (offset === 4 || offset === 6 || offset === 8)
                        result += "-";
                    if (byte < 16)
                        result += "0";
                    result += byte.toString(16).toLowerCase();
                }
                return result;
            }
        }
        // uses a heuristic used by v8 and chakra to force an object into dictionary mode.
        function MakeDictionary(obj) {
            obj.__ = undefined;
            delete obj.__;
            return obj;
        }
    });
})(Reflect || (Reflect = {}));

class ConnectionManager {
  /**
   * List of connections registered in this connection manager.
   *
   * @public
   * @returns {DataSource[]} - List of connections
   *
   * @connectionmanager
   */
  get connections() {
    return Array.from(this.connectionMap.values());
  }
  /**
   * Internal lookup to quickly get from a connection name to the Connection object.
   */
  connectionMap = /* @__PURE__ */ new Map();
  // -------------------------------------------------------------------------
  // Public Methods
  // -------------------------------------------------------------------------
  /**
   * Checks if connection with the given name exist in the manager.
   *
   * @public
   * @param {string} name - Connection name
   * @returns {boolean} - True if connection exist, false otherwise
   *
   * @connectionmanager
   */
  has(name) {
    return this.connectionMap.has(name);
  }
  /**
   * Gets registered connection with the given name.
   * If connection name is not given then it will get a default connection.
   * Throws error if connection with the given name was not found.
   *
   * @public
   * @param {string} name - Connection name
   * @returns {DataSource} - Connection
   *
   * @connectionmanager
   */
  get(name = "default") {
    const connection = this.connectionMap.get(name);
    if (!connection)
      throw new typeorm.ConnectionNotFoundError(name);
    return connection;
  }
  /**
   * Removes registered connection with the given name.
   * If connection name is not given then it will get a default connection.
   * Throws error if connection with the given name was not found.
   *
   * @public
   * @param {string} name - Connection name
   *
   * @connectionmanager
   */
  remove(name = "default") {
    const connection = this.connectionMap.get(name);
    if (!connection)
      throw new typeorm.ConnectionNotFoundError(name);
    this.connectionMap.delete(name);
  }
  /**
   * closes registered connection with the given name and removes it from
   * ConnectionManager.
   * If connection name is not given then it will get a default connection.
   * Throws error if connection with the given name was not found.
   *
   * @public
   * @param {string | Array<string>} name - Connection name
   *
   * @connectionmanager
   */
  async close(name = "default") {
    const throwError = (name2) => {
      throw new typeorm.ConnectionNotFoundError(name2);
    };
    const closeConnection = async (name2) => {
      const connection = this.connectionMap.get(name2);
      await connection.destroy();
      this.remove(name2);
    };
    return !lodash.isArray(name) && this.connectionMap.has(name) ? await closeConnection(name).then(() => {
      return true;
    }).catch(() => {
      return false;
    }) : lodash.isArray(name) ? name.map(async (connectionName) => {
      return this.connectionMap.has(connectionName) ? await closeConnection(connectionName).then(() => {
        return true;
      }).catch(() => {
        return false;
      }) : throwError(connectionName);
    }) : throwError(name);
  }
  /**
   * Creates a new connection based on the given connection options and registers it in the manager.
   * Connection won't be established, you'll need to manually call connect method to establish connection.
   *
   * @public
   * @param {Object} options - TypeORM data source connection options
   * @param {boolean} newConnection - Toggle to create a new instance of TypeORMDbAdapter.
   * @returns {Promise<connection>} - Connection
   *
   * @connectionmanager
   */
  async create(options, newConnection = false) {
    const existConnection = this.connectionMap.get(options.name || "default");
    const throwError = () => {
      const error = new typeorm.AlreadyHasActiveConnectionError(options.name || "default");
      throw new moleculer.Errors.MoleculerServerError(
        error.message,
        500,
        "ERR_CONNECTION_ALREADY_EXIST"
      );
    };
    let activeConneciton;
    if (newConnection && !existConnection) {
      return new TypeORMDbAdapter(options);
    } else {
      const dbConnection = existConnection && existConnection.isInitialized ? throwError() : new typeorm.DataSource(options);
      activeConneciton = await dbConnection.initialize().then((dataConnection) => dataConnection).catch((err) => {
        throw new moleculer.Errors.MoleculerServerError(
          err.message,
          500,
          "ERR_CONNECTION_CREATE"
        );
      });
      this.connectionMap.set(dbConnection.name, dbConnection);
    }
    return bluebird.resolve(activeConneciton);
  }
}

class TypeORMDbAdapter {
  /**
   * Grants access to the connection manager instance which is used to create and manage connections.
   * Called using this.adapter.connectionManager
   *
   * @static
   * @property {ConnectionManager} connectionManager
   *
   * @properties
   */
  connectionManager;
  _entity;
  dataSource;
  /**
   * Grants access to the entity manager of the connection.
   * Called using this.adapter.manager
   * @static
   * @property {EntityManager} manager
   *
   * @properties
   */
  manager;
  /**
   * Grants access to the entity repository of the connection.
   * Called using this.adapter.repository
   * @static
   * @property {Repository<Entity>} repository
   *
   * @properties
   */
  repository;
  /**
   * Creates an instance of TypeORM db service.
   *
   * @param {DataSourceOptions} opts
   *
   */
  constructor(opts) {
    this.opts = opts;
  }
  /**
   * Initialize adapter.
   * It will be called in `broker.start()` and is used internally.
   * @methods
   * @param {ServiceBroker} broker
   * @param {Service} service
   * @memberof TypeORMDbAdapter
   */
  init(broker, service) {
    this.broker = broker;
    this.service = service;
    const entityFromService = this.service.schema.model;
    const entityArray = [];
    lodash.has(this.opts, "entities") ? this._entity = [...this.opts.entities] : lodash.isArray(entityFromService) ? (entityFromService.forEach((entity) => {
      const isValid = !!entity.constructor;
      if (!isValid) {
        new moleculer.Errors.MoleculerServerError(
          "Invalid model. It should be a typeorm repository"
        );
      }
      entityArray.push(entity);
    }), this._entity = entityArray) : !lodash.isUndefined(entityFromService) && !!entityFromService.constructor ? this._entity = entityFromService : new moleculer.Errors.MoleculerServerError("Invalid model. It should be a typeorm repository");
  }
  /**
   * Connects to database.
   * It will be called in `broker.start()` and is used internally.
   * @methods
   * @public
   * @returns {Promise}
   */
  async connect() {
    this.connectionManager = new ConnectionManager();
    const db = await this.connectionManager.create(this.opts);
    this.broker.logger.info(`${this.service.name} has connected to ${db.name} database`);
    const entityArrray = lodash.isArray(this._entity) ? this._entity : [this._entity];
    const entityMethods = (obj = {}) => {
      const members = Object.getOwnPropertyNames(obj);
      const methods = members.filter((el) => {
        return typeof obj[el] === "function";
      });
      return methods;
    };
    entityArrray.forEach((entity, index) => {
      const dbRepository = db.getRepository(entity);
      const entityName = dbRepository.metadata.name;
      const methodNames = entityMethods(entity);
      const methodsToAdd = {
        manager: dbRepository.manager,
        repository: dbRepository,
        getRepository: function getRepository() {
          const dataSource = db;
          if (!dataSource)
            throw new Error(`DataSource is not set for this entity.`);
          return dataSource.getRepository(entity);
        }
      };
      methodNames.forEach((method2) => {
        index === 0 ? this[method2] = entity[method2] : methodsToAdd[method2] = entity[method2];
      });
      [
        "hasId",
        "save",
        "remove",
        "softRemove",
        "recover",
        "reload",
        "useDataSource",
        // 'getRepository', // causing issue with typeormdbadapter class getRepository
        "target",
        "getId",
        "createQueryBuilder",
        "create",
        "merge",
        "preload",
        "insert",
        "update",
        "upsert",
        "delete",
        "count",
        "countBy",
        "sum",
        "average",
        "minimum",
        "maximum",
        "find",
        "findBy",
        "findAndCount",
        "findAndCountBy",
        "findOne",
        "findOneBy",
        "findOneOrFail",
        "findOneByOrFail",
        "query",
        "clear"
      ].forEach((method2) => {
        index === 0 ? this[method2] = entity[method2] : methodsToAdd[method2] = entity[method2];
      });
      !entity["save"] ? this.broker.logger.warn(
        `Entity class ${entityName} does not extend TypeORM BaseEntity, use data mapping with this.adapter.repository instead of active record methodology.`
      ) : index !== 0 ? this[entityName] = {
        ...methodsToAdd,
        findByIdWO: this.findByIdWO,
        findById: this.findById,
        findByIds: this.findByIds,
        list: this.list,
        afterRetrieveTransformID: this.afterRetrieveTransformID,
        encodeID: this.encodeID,
        decodeID: this.decodeID,
        transformDocuments: this.transformDocuments,
        beforeEntityChange: this.beforeEntityChange,
        entityChanged: this.entityChanged,
        clearCache: this.clearCache,
        filterFields: this.filterFields,
        excludeFields: this.excludeFields,
        _excludeFields: this._excludeFields,
        populateDocs: this.populateDocs,
        validateEntity: this.validateEntity,
        entityToObject: this.entityToObject,
        beforeSaveTransformID: this.beforeSaveTransformID,
        beforeQueryTransformID: this.beforeQueryTransformID,
        authorizeFields: this.authorizeFields,
        updateById: this.updateById,
        broker: this.broker,
        service: this.service
      } : null;
    });
    this.manager = db.manager;
    this.repository = db.getRepository(lodash.isArray(this._entity) ? this._entity[0] : this._entity);
    this.dataSource = db;
  }
  /**
   * Disconnects all connections from database and connection manager.
   * It will be called in `broker.stop()` and is used internally.
   * @methods
   * @public
   * @returns {Promise}
   * @memberof TypeORMDbAdapter
   */
  disconnect() {
    this.connectionManager.connections.forEach(async (connection) => {
      this.broker.logger.info(`Attempting to disconnect from database ${connection.name}...`);
      await this.connectionManager.close(connection.name).then(
        (disconnected) => disconnected === true ? this.broker.logger.info(`Disconnected from database ${connection.name}`) : this.broker.logger.info(
          `Failed to disconnect from database ${connection.name}`
        )
      ).catch((error) => {
        this.broker.logger.error(`Failed to disconnect from database ${error}`);
        new moleculer.Errors.MoleculerServerError(
          `Failed to disconnect from database ${error}`,
          500,
          "FAILED_TO_DISCONNECT_FROM_DATABASE",
          error
        );
      });
    });
    return bluebird.resolve();
  }
  // -------------------------------------------------------------------------
  // Public Methods
  // -------------------------------------------------------------------------
  /**
   * Gets current entity's Repository and returns it.
   * Needed for active record to work from base entity and
   * Uses this.entity which could be an entity or an array of entities.
   * If this._entity is an array, uses first entity in array for active record.
   * Used internally by this.adapter for base conection.
   * @methods
   * @private
   * @param {T} this
   * @returns {Repository<Entity>}
   * @memberof TypeORMDbAdapter
   *
   */
  getRepository() {
    const dataSource = this.dataSource;
    if (!dataSource)
      throw new Error(`DataSource is not set for this entity.`);
    return dataSource.getRepository(lodash.isArray(this._entity) ? this._entity[0] : this._entity);
  }
  /**
   * Gets item by id. Can use find options
   * @methods
   * @param {Partial<T>} key - primary column name
   * @param {string | number} id - id of entity
   * @param {Object} findOptions - find options, like relations, order, etc. No where clause
   * @returns {Promise<T | undefined>}
   * @memberof TypeORMDbAdapter
   */
  async findByIdWO(key = this.service.settings.idField, id, findOptions) {
    const dbID = this.beforeQueryTransformID(key);
    const entity = await this["findOneOrFail"]({
      where: { [dbID]: typeorm.In([id]) },
      ...findOptions
    });
    return this.afterRetrieveTransformID(entity, this.service.settings.idField);
  }
  /**
   * Gets item by id. No find options
   * @methods
   * @param {Partial<T>} key - primary column name
   * @param {string | number} id - id of entity
   * @returns {Promise<T | undefined>}
   * @memberof TypeORMDbAdapter
   *
   */
  async findById(key = this.service.settings.idField, id) {
    const dbID = this.beforeQueryTransformID(key);
    const entity = await this["findOneByOrFail"]({ [dbID]: typeorm.In([id]) });
    return this.afterRetrieveTransformID(entity, this.service.settings.idField);
  }
  /**
   * Gets items by id.
   * @methods
   * @param {Partial<T>} key - primary column name
   * @param {Array<string> | Array<number>} ids - ids of entity
   * @returns {Promise<T | undefined>}
   * @memberof TypeORMDbAdapter
   *
   */
  async findByIds(key = this.service.settings.idField, ids) {
    const dbID = this.beforeQueryTransformID(key);
    const entity = await this["findBy"]({ [dbID]: typeorm.In([...ids]) });
    return this.afterRetrieveTransformID(entity, this.service.settings.idField);
  }
  /**
   * List entities by filters and pagination results.
   * @methods
   * @param {Context} ctx - Context instance.
   * @param {FindManyOptions<Object>?} params - Optional parameters.
   *
   * @returns {Object} List of found entities and count.
   * @memberof TypeORMDbAdapter
   */
  list(ctx, params) {
    let countParams = Object.assign({}, params);
    if (countParams && countParams.limit)
      countParams.limit = null;
    if (countParams && countParams.offset)
      countParams.offset = null;
    if (params.limit == null) {
      if (this.service.settings.limit > 0 && params.pageSize > this.service.settings.limit)
        params.limit = this.service.settings.limit;
      else
        params.limit = params.pageSize;
    }
    let modifiedFindParams;
    if (params.take && params.skip) {
      if (params.limit) {
        delete params.limit;
      }
      if (params.offset) {
        delete params.offset;
      }
      modifiedFindParams = params;
    } else if (params.limit || params.offset) {
      modifiedFindParams = JSON.parse(
        lodash.replace(JSON.stringify(params), '"limit":', '"take":').replace(
          '"offset":',
          '"skip":'
        )
      );
    }
    if (lodash.has(modifiedFindParams, "relations") && typeof modifiedFindParams.relations === "string") {
      modifiedFindParams.relations = JSON.parse(modifiedFindParams.relations);
    }
    if (lodash.has(modifiedFindParams, "where") && typeof modifiedFindParams.where === "string") {
      modifiedFindParams.where = JSON.parse(modifiedFindParams.where);
    }
    let modifiedCountParams;
    if (countParams.take && countParams.skip) {
      if (countParams.limit) {
        delete countParams.limit;
      }
      if (countParams.offset) {
        delete countParams.offset;
      }
      modifiedCountParams = countParams;
    }
    if (params.limit || params.offset) {
      modifiedCountParams = JSON.parse(
        lodash.replace(JSON.stringify(countParams), '"limit":', '"take":').replace(
          '"offset":',
          '"skip":'
        )
      );
    }
    if (lodash.has(modifiedCountParams, "relations")) {
      delete modifiedCountParams.relations;
    }
    if (lodash.has(modifiedCountParams, "where") && typeof modifiedCountParams.where === "string") {
      modifiedCountParams.where = JSON.parse(modifiedCountParams.where);
    }
    delete modifiedFindParams.pageSize;
    delete modifiedFindParams.page;
    delete modifiedCountParams.pageSize;
    delete modifiedCountParams.page;
    console.log("list params: ", modifiedFindParams);
    console.log("count params: ", countParams);
    console.log("params: ", params);
    console.log("modified modifiedCountParams: ", modifiedCountParams);
    return bluebird.all([
      // Get rows
      this.find(modifiedFindParams),
      // Get count of all rows
      this.count(modifiedCountParams)
    ]).then(async (res) => {
      return await this.transformDocuments(ctx, params, res[0]).then((docs) => {
        return {
          // Rows
          rows: docs,
          // Total rows
          total: res[1],
          // Page
          page: params.page,
          // Page size
          pageSize: params.pageSize,
          // Total pages
          totalPages: Math.floor((res[1] + params.pageSize - 1) / params.pageSize)
        };
      });
    });
  }
  /**
   * Transforms 'idField' into expected db id field.
   * @methods
   * @param {Object} entity
   * @param {String} idField
   * @memberof MemoryDbAdapter
   * @returns {Object} Modified entity
   * @memberof TypeORMDbAdapter
   */
  beforeSaveTransformID(entity, idField) {
    let newEntity = lodash.cloneDeep(entity);
    const dbIDField = lodash.find(this.repository.metadata.ownColumns, {
      isPrimary: true
    }).propertyName;
    if (idField !== dbIDField && entity[idField] !== void 0) {
      newEntity = JSON.parse(
        lodash.replace(
          JSON.stringify(newEntity),
          new RegExp(`"${idField}":`, "g"),
          `"${dbIDField}":`
        )
      );
    }
    return newEntity;
  }
  /**
   * Transforms db field into user defined 'idField'
   * @methods
   * @param {Object} entity
   * @param {String} idField
   * @memberof MemoryDbAdapter
   * @returns {Object} Modified entity
   * @memberof TypeORMDbAdapter
   */
  afterRetrieveTransformID(entity, idField) {
    const dbIDField = lodash.find(this.repository.metadata.ownColumns, {
      isPrimary: true
    }).propertyName;
    let newEntity;
    if (!entity.hasOwnProperty(idField)) {
      newEntity = JSON.parse(
        lodash.replace(
          JSON.stringify(entity),
          new RegExp(`"${dbIDField}":`, "g"),
          `"${idField}":`
        )
      );
    }
    return newEntity;
  }
  /**
   * Encode ID of entity.
   * @methods
   * @param {any} id
   * @returns {any}
   * @memberof TypeORMDbAdapter
   */
  encodeID(id) {
    return id;
  }
  /**
   * Transform idField into the name of the id field in db
   * @methods
   * @param {any} idField
   * @returns {any}
   * @memberof TypeORMDbAdapter
   */
  beforeQueryTransformID(idField) {
    const dbIDField = lodash.find(this.repository.metadata.ownColumns, {
      isPrimary: true
    }).propertyName;
    if (idField !== dbIDField) {
      return dbIDField;
    }
    return idField;
  }
  /**
   * Decode ID of entity.
   * @methods
   * @param {any} id
   * @returns {any}
   * @memberof TypeORMDbAdapter
   */
  decodeID(id) {
    return id;
  }
  /**
   * Transform the fetched documents
   * @methods
   * @param {Context} ctx
   * @param {Object} 	params
   * @param {Array|Object} docs
   * @returns {Array|Object}
   * @memberof TypeORMDbAdapter
   */
  transformDocuments(ctx, params, docs) {
    let isDoc = false;
    if (!Array.isArray(docs)) {
      if (lodash.isObject(docs)) {
        isDoc = true;
        docs = [docs];
      } else
        return bluebird.resolve(docs);
    }
    return bluebird.resolve(docs).then((docs2) => docs2.map((doc) => this.entityToObject(doc))).then(
      (docs2) => docs2.map(
        (doc) => this.afterRetrieveTransformID(doc, this.service.settings.idField)
      )
    ).then(
      (docs2) => docs2.map((doc) => {
        doc[this.service.settings.idField] = this.encodeID(
          doc[this.service.settings.idField]
        );
        return doc;
      })
    ).then(
      (json) => ctx && params.populate ? this.populateDocs(ctx, json, params.populate) : json
    ).then((json) => {
      if (ctx && params.fields) {
        const fields = lodash.isString(params.fields) ? (
          // Compatibility with < 0.4
          /* istanbul ignore next */
          params.fields.split(/\s+/)
        ) : params.fields;
        const authFields = this.authorizeFields(fields);
        return json.map((item) => this.filterFields(item, authFields));
      } else {
        return json.map(
          (item) => this.filterFields(item, this.service.settings.fields)
        );
      }
    }).then((json) => {
      const askedExcludeFields = ctx && params.excludeFields ? lodash.isString(params.excludeFields) ? params.excludeFields.split(/\s+/) : params.excludeFields : [];
      const excludeFields = askedExcludeFields.concat(
        this.service.settings.excludeFields || []
      );
      if (Array.isArray(excludeFields) && excludeFields.length > 0) {
        return json.map((doc) => {
          return this._excludeFields(doc, excludeFields);
        });
      } else {
        return json;
      }
    }).then((json) => isDoc ? json[0] : json).catch((err) => {
      this.broker.logger.error("Transforming documents is failed!", err);
      throw new moleculer.Errors.MoleculerServerError(
        `Failed to transform documents ${err}`,
        500,
        "FAILED_TO_TRANSFORM_DOCUMENTS",
        err
      );
    });
  }
  /**
   * Call before entity lifecycle events
   * @methods
   * @param {String} type
   * @param {Object} entity
   * @param {Context} ctx
   * @returns {Promise}
   * @memberof TypeORMDbAdapter
   */
  beforeEntityChange(type, entity, ctx) {
    const eventName = `beforeEntity${lodash.capitalize(type)}`;
    if (this.service.schema[eventName] == null) {
      return bluebird.resolve(entity);
    }
    return bluebird.resolve(this.service.schema[eventName].call(this, entity, ctx));
  }
  /**
   * Clear the cache & call entity lifecycle events
   * @methods
   * @param {String} type
   * @param {Object|Array<Object>|Number} json
   * @param {Context} ctx
   * @returns {Promise}
   * @memberof TypeORMDbAdapter
   */
  async entityChanged(type, json, ctx) {
    return await this.clearCache().then(async () => {
      const eventName = `entity${lodash.capitalize(type)}`;
      if (this.schema[eventName] != null) {
        return await this.service.schema[eventName].call(this, json, ctx);
      }
    });
  }
  /**
   * Clear cached entities
   * @methods
   * @returns {Promise}
   * @memberof TypeORMDbAdapter
   */
  clearCache() {
    this.broker[this.service.settings.cacheCleanEventType](`cache.clean.${this.fullName}`);
    if (this.broker.cacher)
      return this.broker.cacher.clean(`${this.fullName}.**`);
    return bluebird.resolve();
  }
  /**
   * Filter fields in the entity object
   * @methods
   * @param {Object} 	doc
   * @param {Array<String>} 	fields	Filter properties of model.
   * @returns	{Object}
   * @memberof TypeORMDbAdapter
   */
  filterFields(doc, fields) {
    if (Array.isArray(fields)) {
      let res = {};
      fields.forEach((n) => {
        const v = lodash.get(doc, n);
        if (v !== void 0)
          lodash.set(res, n, v);
      });
      return res;
    }
    return doc;
  }
  /**
   * Exclude fields in the entity object
   * @methods
   * @param {Object} 	doc
   * @param {Array<String>} 	fields	Exclude properties of model.
   * @returns	{Object}
   * @memberof TypeORMDbAdapter
   */
  excludeFields(doc, fields) {
    if (Array.isArray(fields) && fields.length > 0) {
      return this._excludeFields(doc, fields);
    }
    return doc;
  }
  /**
   * Exclude fields in the entity object. Internal use only, must ensure `fields` is an Array
   */
  _excludeFields(doc, fields) {
    const res = lodash.cloneDeep(doc);
    fields.forEach((field) => {
      lodash.unset(res, field);
    });
    return res;
  }
  /**
   * Populate documents.
   * @methods
   * @param {Context} 		ctx
   * @param {Array|Object} 	docs
   * @param {Array?}			populateFields
   * @returns	{Promise}
   * @memberof TypeORMDbAdapter
   */
  populateDocs(ctx, docs, populateFields) {
    if (!this.service.settings.populates || !Array.isArray(populateFields) || populateFields.length == 0)
      return bluebird.resolve(docs);
    if (docs == null || !lodash.isObject(docs) && !Array.isArray(docs))
      return bluebird.resolve(docs);
    const settingPopulateFields = Object.keys(this.service.settings.populates);
    const groupedPopulateFields = populateFields.reduce((obj, populateField) => {
      const settingPopulateField = settingPopulateFields.find(
        (settingPopulateField2) => settingPopulateField2 === populateField || populateField.startsWith(settingPopulateField2 + ".")
      );
      if (settingPopulateField != null) {
        if (obj[settingPopulateField] == null) {
          obj[settingPopulateField] = [populateField];
        } else {
          obj[settingPopulateField].push(populateField);
        }
      }
      return obj;
    }, {});
    let promises = [];
    for (const populatesField of settingPopulateFields) {
      let rule = this.service.settings.populates[populatesField];
      if (groupedPopulateFields[populatesField] == null)
        continue;
      if (lodash.isFunction(rule)) {
        rule = {
          handler: bluebird.method(rule)
        };
      }
      if (lodash.isString(rule)) {
        rule = {
          action: rule
        };
      }
      if (rule.field === void 0)
        rule.field = populatesField;
      let arr = Array.isArray(docs) ? docs : [docs];
      let idList = lodash.uniq(lodash.flattenDeep(lodash.compact(arr.map((doc) => lodash.get(doc, rule.field)))));
      const resultTransform = (populatedDocs) => {
        arr.forEach((doc) => {
          let id = lodash.get(doc, rule.field);
          if (lodash.isArray(id)) {
            let models = lodash.compact(id.map((id2) => populatedDocs[id2]));
            lodash.set(doc, populatesField, models);
          } else {
            lodash.set(doc, populatesField, populatedDocs[id]);
          }
        });
      };
      if (rule.handler) {
        promises.push(rule.handler.call(this, idList, arr, rule, ctx));
      } else if (idList.length > 0) {
        const params = Object.assign(
          {
            id: idList,
            mapping: true,
            populate: [
              // Transform "post.author" into "author" to pass to next populating service
              ...groupedPopulateFields[populatesField].map(
                (populateField) => populateField.slice(populatesField.length + 1)
              ).filter((field) => field !== ""),
              ...rule.populate ? rule.populate : []
            ]
          },
          rule.params || {}
        );
        if (params.populate.length === 0) {
          delete params.populate;
        }
        promises.push(ctx.call(rule.action, params).then(resultTransform));
      }
    }
    return bluebird.all(promises).then(() => docs);
  }
  /**
   * Validate an entity by validator.
   * @methods
   * @param {Object} entity
   * @returns {Promise}
   * @memberof TypeORMDbAdapter
   */
  validateEntity(entity) {
    if (!lodash.isFunction(this.service.settings.entityValidator))
      return bluebird.resolve(entity);
    let entities = Array.isArray(entity) ? entity : [entity];
    return bluebird.all(
      entities.map((entity2) => this.service.settings.entityValidator.call(this, entity2))
    ).then(() => entity);
  }
  /**
   * Convert DB entity to JSON object
   * @methods
   * @param {any} entity
   * @returns {Object}
   * @memberof TypeORMDbAdapter
   */
  entityToObject(entity) {
    return entity;
  }
  /**
   * Authorize the required field list. Remove fields which is not exist in the `this.service.settings.fields`
   * @methods
   * @param {Array} askedFields
   * @returns {Array}
   * @memberof TypeORMDbAdapter
   */
  authorizeFields(askedFields) {
    if (this.service.settings.fields && this.service.settings.fields.length > 0) {
      let allowedFields = [];
      if (Array.isArray(askedFields) && askedFields.length > 0) {
        askedFields.forEach((askedField) => {
          if (this.service.settings.fields.indexOf(askedField) !== -1) {
            allowedFields.push(askedField);
            return;
          }
          if (askedField.indexOf(".") !== -1) {
            let parts = askedField.split(".");
            while (parts.length > 1) {
              parts.pop();
              if (this.service.settings.fields.indexOf(parts.join(".")) !== -1) {
                allowedFields.push(askedField);
                return;
              }
            }
          }
          let nestedFields = this.service.settings.fields.filter(
            (settingField) => settingField.startsWith(askedField + ".")
          );
          if (nestedFields.length > 0) {
            allowedFields = allowedFields.concat(nestedFields);
          }
        });
      }
      return allowedFields;
    }
    return askedFields;
  }
  /**
   * Update an entity by ID
   * @methods
   * @param {any} id
   * @param {Object} update
   * @returns {Promise}
   * @memberof TypeORMDbAdapter
   */
  async updateById(id, update) {
    const entity = await this["update"]({ id }, update);
    return this.afterRetrieveTransformID(entity, this.service.settings.idField);
  }
  /**
   * Sanitize context parameters at `find` action.
   *
   * @methods
   *
   * @param {Context} ctx
   * @param {Object} params
   * @returns {Object}
   */
  sanitizeParams(ctx, params) {
    let p = Object.assign({}, params);
    if (typeof p.limit === "string")
      p.limit = Number(p.limit);
    if (typeof p.offset === "string")
      p.offset = Number(p.offset);
    if (typeof p.page === "string")
      p.page = Number(p.page);
    if (typeof p.pageSize === "string")
      p.pageSize = Number(p.pageSize);
    if (typeof p.query === "string")
      p.query = JSON.parse(p.query);
    if (typeof p.sort === "string")
      p.sort = p.sort.split(/[,\s]+/);
    if (typeof p.fields === "string")
      p.fields = p.fields.split(/[,\s]+/);
    if (typeof p.excludeFields === "string")
      p.excludeFields = p.excludeFields.split(/[,\s]+/);
    if (typeof p.populate === "string")
      p.populate = p.populate.split(/[,\s]+/);
    if (typeof p.searchFields === "string")
      p.searchFields = p.searchFields.split(/[,\s]+/);
    if (ctx.action.name.endsWith(".list")) {
      if (!p.pageSize)
        p.pageSize = this.settings.pageSize;
      if (!p.page)
        p.page = 1;
      if (this.settings.maxPageSize > 0 && p.pageSize > this.settings.maxPageSize)
        p.pageSize = this.settings.maxPageSize;
      p.limit = p.pageSize;
      p.offset = (p.page - 1) * p.pageSize;
    }
    if (this.settings.maxLimit > 0 && p.limit > this.settings.maxLimit)
      p.limit = this.settings.maxLimit;
    return p;
  }
}

module.exports = TypeORMDbAdapter;
